@@ -1,15 +1,22 @@
 
 
 import argparse
+from asyncio.log import logger
+from dataclasses import dataclass
+from enum import IntEnum
+from typing import TYPE_CHECKING, Optional
 import liblo
 import os
 import shlex
 import socket
 import subprocess
 import sys
+import logging
+
 from liblo import Server, Address
-from PyQt5.QtCore import QT_VERSION_STR, QFile
-from PyQt5.QtGui import QIcon, QPalette
+from PyQt5.QtCore import QT_VERSION_STR, QSettings
+
+_logger = logging.getLogger(__name__)
 
 # get qt version in list of ints
 QT_VERSION = []
@@ -23,7 +30,7 @@ if QT_VERSION < (5, 6):
         "WARNING: You are using a version of QT older than 5.6.\n"
         + "You won't be warned if a process can't be launch.\n")
 
-VERSION = "0.12.2"
+VERSION = "0.14.2"
 
 APP_TITLE = 'RaySession'
 DEFAULT_SESSION_ROOT = "%s/Ray Sessions" % os.getenv('HOME')
@@ -35,8 +42,7 @@ RAYNET_BIN = 'ray-network'
 
 GIT_IGNORED_EXTENSIONS = ".wav .flac .ogg .mp3 .mp4 .avi .mkv .peak .m4a .pdf"
 
-GROUP_CONTEXT_AUDIO = 0x01
-GROUP_CONTEXT_MIDI = 0x02
+
 
 class PrefixMode:
     CUSTOM = 0
@@ -63,6 +69,7 @@ class ClientStatus:
     REMOVED = 11
     UNDEF = 12
     SCRIPT = 13
+    LOSE = 14
 
 
 class ServerStatus:
@@ -154,7 +161,7 @@ class WaitFor:
     SCRIPT_QUIT = 7
 
 
-class Template:
+class Template(IntEnum):
     NONE = 0
     RENAME = 1
     SESSION_SAVE = 2
@@ -183,12 +190,12 @@ class Systray:
     SESSION_ONLY = 1
     ALWAYS = 2
 
-
-class Favorite():
-    def __init__(self, name, icon, factory):
-        self.name = name
-        self.icon = icon
-        self.factory = factory
+@dataclass
+class Favorite:
+    name: str
+    icon: str
+    factory: bool
+    display_name: str
 
 
 class ScriptFile:
@@ -199,7 +206,7 @@ class ScriptFile:
     CLOSE = 0x8
 
     @classmethod
-    def by_string(cls, action:str)->int:
+    def by_string(cls, action: str) -> int:
         if action == 'load':
             return cls.LOAD
         if action == 'save':
@@ -215,7 +222,7 @@ def ifDebug(string):
     if debug:
         sys.stderr.write("%s\n" % string)
 
-def version_to_tuple(version_str):
+def version_to_tuple(version_str: str):
     version_list = []
     for c in version_str.split('.'):
         if not c.isdigit():
@@ -225,14 +232,14 @@ def version_to_tuple(version_str):
     return tuple(version_list)
 
 def add_self_bin_to_path():
-    # Add raysession/src/bin to $PATH to can use ray executables after make
+    # Add RaySession/src/bin to $PATH to can use ray executables after make
     # Warning, will works only if link to this file is in RaySession/*/*/*.py
     this_path = os.path.realpath(os.path.dirname(os.path.realpath(__file__)))
     bin_path = "%s/bin" % os.path.dirname(this_path)
     if not os.environ['PATH'].startswith("%s:" % bin_path):
         os.environ['PATH'] = "%s:%s" % (bin_path, os.environ['PATH'])
 
-def get_list_in_settings(settings, path):
+def get_list_in_settings(settings: QSettings, path: str) -> list:
     # getting a QSettings value of list type seems to not works the same way
     # on all machines
     try:
@@ -243,9 +250,11 @@ def get_list_in_settings(settings, path):
         except BaseException:
             settings_list = []
 
-    return settings_list
+    if isinstance(settings_list, list):
+        return settings_list
+    return []
 
-def is_git_taggable(string)->bool:
+def is_git_taggable(string: str) -> bool:
     ''' know if a string can be a git tag, not used currently '''
     if not string:
         return False
@@ -269,7 +278,7 @@ def is_git_taggable(string)->bool:
 
     return True
 
-def is_valid_full_path(path: str)->bool:
+def is_valid_full_path(path: str) -> bool:
     if not path.startswith('/'):
         return False
 
@@ -281,7 +290,7 @@ def is_valid_full_path(path: str)->bool:
         return False
     return True
 
-def is_osc_port_free(port:int):
+def is_osc_port_free(port: int) -> bool:
     try:
         testport = Server(port)
     except BaseException:
@@ -312,14 +321,14 @@ def get_free_osc_port(default=16187):
     del testport
     return daemon_port
 
-def is_valid_osc_url(url:str):
+def is_valid_osc_url(url: str) -> bool:
     try:
         address = liblo.Address(url)
         return True
     except BaseException:
         return False
 
-def get_liblo_address(url):
+def get_liblo_address(url: str) -> Optional[liblo.Address]:
     valid_url = False
     try:
         address = liblo.Address(url)
@@ -337,7 +346,7 @@ def get_liblo_address(url):
             msg = "%r is an unknown osc url" % url
             raise argparse.ArgumentTypeError(msg)
 
-def get_liblo_address_from_port(port:int):
+def get_liblo_address_from_port(port:int) -> Optional[liblo.Address]:
     try:
         port = int(port)
     except:
@@ -430,14 +439,14 @@ def are_on_same_machine(url1, url2):
 
     return False
 
-def get_net_url(port)->str:
+def get_net_url(port) -> str:
     ip = Machine192.get()
     if not ip:
         return ''
 
     return "osc.udp://%s:%i/" % (ip, port)
 
-def shell_line_to_args(string:str)->list:
+def shell_line_to_args(string:str) -> list:
     try:
         args = shlex.split(string)
     except BaseException:
@@ -445,19 +454,19 @@ def shell_line_to_args(string:str)->list:
 
     return args
 
-def types_are_all_strings(types:str)->bool:
+def types_are_all_strings(types:str) -> bool:
     for char in types:
         if char != 's':
             return False
     return True
 
-def are_they_all_strings(args:list)->bool:
+def are_they_all_strings(args:list) -> bool:
     for arg in args:
         if not isinstance(arg, str):
             return False
     return True
 
-def get_window_manager():
+def get_window_manager() -> WindowManager:
     if os.getenv('WAYLAND_DISPLAY'):
         return WindowManager.WAYLAND
 
@@ -466,25 +475,14 @@ def get_window_manager():
 
     return WindowManager.NONE
 
-def get_full_path(root, session_name):
-    spath = "%s%s%s" % (root, os.sep, session_name)
-
-    if session_name.startswith(os.sep):
-        spath = session_name
-
-    if spath.endswith(os.sep):
-        spath = spath[:-1]
-
-    return spath
-
-def protocol_to_str(protocol: int)->str:
+def protocol_to_str(protocol: int) -> str:
     if protocol == Protocol.RAY_HACK:
         return "Ray-Hack"
     if protocol == Protocol.RAY_NET:
         return "Ray-Net"
     return "NSM"
 
-def protocol_from_str(protocol_str: str)->int:
+def protocol_from_str(protocol_str: str) -> int:
     if protocol_str.lower() in ('ray_hack', 'ray-hack'):
         return Protocol.RAY_HACK
     elif protocol_str.lower() in ('ray_net', 'ray-net'):
@@ -497,7 +495,7 @@ class Machine192:
     read_done = False
     
     @staticmethod
-    def read()->str:
+    def read() -> str:
         try:
             ips = subprocess.check_output(
                 ['ip', 'route', 'get', '1']).decode()
@@ -547,12 +545,13 @@ class ClientData:
     template_origin = ''
     jack_client_name = ''
     jack_naming = 0
-    ray_hack = None
-    ray_net = None
+    in_terminal = False
+    ray_hack: 'RayHack' = None
+    ray_net: 'RayNet' = None
 
     @staticmethod
     def sisi():
-        return 'sissssissssssisssi'
+        return 'sissssissssssisssii'
 
     @staticmethod
     def new_from(*args):
@@ -561,7 +560,7 @@ class ClientData:
         return client_data
 
     @staticmethod
-    def spread_client(client)->tuple:
+    def spread_client(client: 'ClientData') -> tuple:
         return (client.client_id, client.protocol,
                 client.executable_path, client.arguments, client.pre_env,
                 client.name, client.prefix_mode, client.custom_prefix,
@@ -570,11 +569,8 @@ class ClientData:
                 client.capabilities, int(client.check_last_save),
                 client.ignored_extensions,
                 client.template_origin,
-                client.jack_client_name, client.jack_naming)
-
-    def gui_init(self, client_id, protocol):
-        self.client_id = client_id
-        self.protocol = protocol
+                client.jack_client_name, client.jack_naming,
+                int(client.in_terminal))
 
     def set_ray_hack(self, ray_hack):
         self.ray_hack = ray_hack
@@ -591,6 +587,7 @@ class ClientData:
                ignored_extensions,
                template_origin,
                jack_client_name, jack_naming,
+               in_terminal,
                secure=False):
         self.executable_path = str(executable)
         self.arguments = str(arguments)
@@ -605,10 +602,13 @@ class ClientData:
         self.ignored_extensions = str(ignored_extensions)
         self.template_origin = template_origin
         self.jack_naming = jack_naming
+        self.in_terminal = bool(in_terminal)
 
         if secure:
             return
 
+        # Now, if message is 'unsecure' only.
+        # change things that can't be changed normally
         self.client_id = str(client_id)
         self.protocol = int(protocol)
         if name:
@@ -629,10 +629,10 @@ class ClientData:
     def update_secure(self, *args):
         self.update(*args, secure=True)
 
-    def spread(self)->tuple:
+    def spread(self) -> tuple:
         return ClientData.spread_client(self)
     
-    def prettier_name(self)->str:
+    def prettier_name(self) -> str:
         if self.label:
             return self.label
         if (self.protocol != Protocol.RAY_HACK
@@ -714,7 +714,7 @@ class RayNet:
 
 
 class GroupPosition:
-    port_types_view = GROUP_CONTEXT_AUDIO | GROUP_CONTEXT_MIDI
+    port_types_view = 3
     group_name = ''
     null_zone = ''
     in_zone = ''
@@ -723,17 +723,19 @@ class GroupPosition:
     in_xy = (0, 0)
     out_xy = (0, 0)
     flags = 0
+    layout_mode = 0
     fully_set = True
     
     @staticmethod
     def get_attributes():
         return ('port_types_view', 'group_name',
                 'null_zone', 'in_zone', 'out_zone',
-                'null_xy', 'in_xy', 'out_xy', 'flags')
+                'null_xy', 'in_xy', 'out_xy', 'flags',
+                'layout_mode')
     
     @staticmethod
     def sisi():
-        return 'issssiiiiiii'
+        return 'issssiiiiiiii'
     
     @staticmethod
     def new_from(*args):
@@ -763,7 +765,7 @@ class GroupPosition:
             
             self.__setattr__(attr, value)
     
-    def is_same(self, other)->bool:
+    def is_same(self, other: 'GroupPosition') -> bool:
         if (self.port_types_view == other.port_types_view
                 and self.group_name == other.group_name):
             return True
@@ -773,17 +775,18 @@ class GroupPosition:
     def update(self, port_types_view: int, group_name: str,
                null_zone: str, in_zone: str, out_zone: str,
                null_x: int, null_y: int, in_x: int, in_y: int,
-               out_x: int, out_y: int, flags: int):
+               out_x: int, out_y: int, flags: int, layout_mode: int):
+        
         for string in (group_name, null_zone, in_zone, out_zone):
-            if type(string) != str:
+            if not isinstance(string, str):
                 return 
         
         for digit in (port_types_view, null_x, null_y, in_x, in_y,
-                      out_x, out_y, flags):
-            if type(digit) == int:
+                      out_x, out_y, flags, layout_mode):
+            if isinstance(digit, int):
                 continue
-            
-            if type(digit) != str:
+
+            if not isinstance(digit, str):
                 return
             
             if (digit.isdigit()
@@ -802,14 +805,16 @@ class GroupPosition:
         self.in_xy = (int(in_x), int(in_y))
         self.out_xy = (int(out_x), int(out_y))
         self.flags = int(flags)
+        self.layout_mode = int(layout_mode)
         
-    def spread(self)->tuple:
+    def spread(self) -> tuple:
         return (self.port_types_view, self.group_name,
                 self.null_zone, self.in_zone, self.out_zone,
                 self.null_xy[0], self.null_xy[1], self.in_xy[0], self.in_xy[1],
-                self.out_xy[0], self.out_xy[1], self.flags)
+                self.out_xy[0], self.out_xy[1], self.flags,
+                self.layout_mode)
     
-    def to_dict(self)->dict:
+    def to_dict(self) -> dict:
         new_dict = {}
         
         for attr in self.__dir__():
@@ -824,11 +829,35 @@ class GroupPosition:
         
         return str(self.__getattribute__(attr))
     
+    def set_layout_mode(self, port_mode: int, layout_mode: int):
+        if not (1 <= port_mode <= 3 or 0 <= layout_mode <= 2):
+            logger.warning(
+                "group position set_layout_mode, wrong port mode or layout mode")
+            return
+
+        layout_mode_str = "%03d" % self.layout_mode
+        new_string = ''
+        for i in range(len(layout_mode_str)):
+            if i == 3 - port_mode:
+                new_string += str(layout_mode)
+            else:
+                new_string += layout_mode_str[i]
+        
+        self.layout_mode = int(new_string)
+        
+    def get_layout_mode(self, port_mode: int) -> int:
+        if not(1 <= port_mode <= 3):
+            return 0
+        
+        layout_mode_str = "%03d" % self.layout_mode
+        return int(layout_mode_str[3 - port_mode])
+
+
 class PortGroupMemory:
     group_name = ''
     port_type = 0
     port_mode = 0
-    port_names = []
+    port_names = list[str]()
     above_metadatas = False
     
     @staticmethod
@@ -883,7 +912,7 @@ class PortGroupMemory:
         
         return new_dict
     
-    def has_a_common_port_with(self, other)->bool:
+    def has_a_common_port_with(self, other: 'PortGroupMemory') -> bool:
         if (self.port_type != other.port_type
                 or self.port_mode != other.port_mode
                 or self.group_name != other.group_name):
